use std::sync::atomic::{AtomicU32, Ordering};
use std::{fmt::Debug, sync::Arc};
use tracing::{
    field::{Field, Visit},
    Subscriber,
};
use tracing_subscriber::{
    layer::{Context, Layer},
    registry::{LookupSpan, SpanRef},
};

#[derive(Clone, Debug)]
struct FieldFilter {
    name: String,
    value: Option<String>,
}

#[derive(Clone, Debug)]
pub struct Filters {
    filter_target: Option<String>,
    filter_name: Option<String>,
    filter_message: Option<String>,
    filter_fields: Vec<FieldFilter>,
}

impl Filters {
    pub fn new() -> Self {
        Self {
            filter_target: None,
            filter_name: None,
            filter_message: None,
            filter_fields: Vec::new(),
        }
    }

    pub fn target(mut self, target: impl ToString) -> Self {
        self.filter_target = Some(target.to_string());
        self
    }

    pub fn name(mut self, name: impl ToString) -> Self {
        self.filter_name = Some(name.to_string());
        self
    }

    pub fn message(mut self, message: impl ToString) -> Self {
        self.filter_message = Some(message.to_string());
        self
    }

    pub fn field(mut self, field: impl ToString) -> Self {
        self.filter_fields.push(FieldFilter {
            name: field.to_string(),
            value: None,
        });
        self
    }

    pub fn field_value(mut self, field: impl ToString, value: impl ToString) -> Self {
        self.filter_fields.push(FieldFilter {
            name: field.to_string(),
            value: Some(value.to_string()),
        });
        self
    }

    fn matches_metadata(&self, metadata: &tracing::Metadata) -> bool {
        if let Some(target) = &self.filter_target {
            if metadata.target() != target {
                return false;
            }
        }
        if let Some(name) = &self.filter_name {
            if metadata.name() != name {
                return false;
            }
        }
        true
    }
}

struct TraceListenerVisitor<'a> {
    filters: &'a Filters,
    right_message: bool,
    right_fields: Vec<bool>,
}

impl<'a> TraceListenerVisitor<'a> {
    fn new(filters: &'a Filters) -> Self {
        Self {
            filters,
            right_message: false,
            right_fields: vec![false; filters.filter_fields.len()],
        }
    }

    fn did_match(&self) -> bool {
        (self.filters.filter_message.is_none() || self.right_message)
            && self.right_fields.iter().all(|right| *right)
    }
}
impl<'a> Visit for TraceListenerVisitor<'a> {
    fn record_debug(&mut self, field: &Field, value: &dyn Debug) {
        use std::fmt::Write;
        let mut value_str = String::new();
        write!(value_str, "{:?}", value).unwrap();
        if let Some(message) = &self.filters.filter_message {
            if field.name() == "message" && &value_str == message {
                self.right_message = true;
            }
        }
        for (filter_field, right) in self
            .filters
            .filter_fields
            .iter()
            .zip(self.right_fields.iter_mut())
        {
            if field.name() == filter_field.name {
                if let Some(filter_field_value) = &filter_field.value {
                    *right = value_str.starts_with(filter_field_value);
                } else {
                    *right = true;
                }
            }
        }
    }
}

#[derive(Clone, Debug)]
pub struct EventListener {
    count: Arc<AtomicU32>,
    filters: Filters,
}

impl EventListener {
    pub fn new(filters: Filters) -> Self {
        Self {
            count: Arc::new(AtomicU32::new(0)),
            filters,
        }
    }

    pub fn get_count(&self) -> u32 {
        self.count.load(Ordering::SeqCst)
    }
}
impl<S: Subscriber> Layer<S> for EventListener {
    fn on_event(&self, ev: &tracing::Event, _: Context<S>) {
        if !self.filters.matches_metadata(ev.metadata()) {
            return;
        }
        let mut visit = TraceListenerVisitor::new(&self.filters);
        ev.record(&mut visit);
        if visit.did_match() {
            self.count.fetch_add(1, Ordering::SeqCst);
        }
    }
}

#[derive(Clone, Debug)]
pub struct SpanFilters {
    filters: Filters,
}

#[allow(dead_code)]
impl SpanFilters {
    pub fn new() -> Self {
        Self {
            filters: Filters::new(),
        }
    }
    pub fn target(mut self, target: impl ToString) -> Self {
        self.filters = self.filters.target(target);
        self
    }

    pub fn name(mut self, name: impl ToString) -> Self {
        self.filters = self.filters.name(name);
        self
    }

    pub fn message(mut self, message: impl ToString) -> Self {
        self.filters = self.filters.message(message);
        self
    }

    pub fn field(mut self, field: impl ToString) -> Self {
        self.filters = self.filters.field(field);
        self
    }

    pub fn field_value(mut self, field: impl ToString, value: impl ToString) -> Self {
        self.filters = self.filters.field_value(field, value);
        self
    }
}

impl SpanFilters {
    fn span_matches<S>(&self, span: &SpanRef<S>) -> bool
    where
        S: Subscriber + for<'lookup> LookupSpan<'lookup>,
    {
        if !self.filters.matches_metadata(span.metadata()) {
            return false;
        }
        true
    }
}

#[derive(Debug)]
struct SpanListenerData {
    enter_count: AtomicU32,
    exit_count: AtomicU32,
}
impl SpanListenerData {
    fn new() -> Self {
        Self {
            enter_count: AtomicU32::new(0),
            exit_count: AtomicU32::new(0),
        }
    }
}
#[derive(Clone, Debug)]
pub struct SpanListener {
    data: Arc<SpanListenerData>,
    filters: SpanFilters,
}

#[allow(dead_code)]
impl SpanListener {
    pub fn new(filters: SpanFilters) -> Self {
        Self {
            data: Arc::new(SpanListenerData::new()),
            filters,
        }
    }

    pub fn get_enter_count(&self) -> u32 {
        self.data.enter_count.load(Ordering::SeqCst)
    }

    pub fn get_exit_count(&self) -> u32 {
        self.data.enter_count.load(Ordering::SeqCst)
    }
}

impl<S: Subscriber + for<'lookup> LookupSpan<'lookup>> Layer<S> for SpanListener {
    fn on_enter(&self, id: &tracing::span::Id, ctx: Context<S>) {
        let span = match ctx.span(id) {
            Some(span) => span,
            None => return,
        };
        if !self.filters.span_matches(&span) {
            return;
        }
        self.data.enter_count.fetch_add(1, Ordering::SeqCst);
    }
    fn on_exit(&self, id: &tracing::span::Id, ctx: Context<S>) {
        let span = match ctx.span(id) {
            Some(span) => span,
            None => return,
        };
        if !self.filters.span_matches(&span) {
            return;
        }
        self.data.exit_count.fetch_add(1, Ordering::SeqCst);
    }
}

mod test {
    use super::*;
    use tracing::info;
    use tracing_subscriber::prelude::*;

    #[test]
    fn test_event_listener() {
        let none_listener = EventListener::new(Filters::new().target("not my module path"));
        let all_listener = EventListener::new(Filters::new().target(std::module_path!()));
        let msg_listener = EventListener::new(Filters::new().message("filter message"));
        let field_listener = EventListener::new(Filters::new().field("field"));
        let field_value_listener = EventListener::new(Filters::new().field_value("field", 1234));
        let msg_field_value_listener = EventListener::new(
            Filters::new()
                .message("filter message")
                .field_value("field", 1234),
        );
        let subscriber = tracing_subscriber::registry()
            .with(all_listener.clone())
            .with(msg_listener.clone())
            .with(field_listener.clone())
            .with(field_value_listener.clone())
            .with(msg_field_value_listener.clone());
        let _sub = tracing::subscriber::set_default(subscriber);

        assert_eq!(none_listener.get_count(), 0);
        assert_eq!(all_listener.get_count(), 0);
        assert_eq!(msg_listener.get_count(), 0);
        assert_eq!(field_listener.get_count(), 0);
        assert_eq!(field_value_listener.get_count(), 0);
        assert_eq!(msg_field_value_listener.get_count(), 0);

        info!("not filter message");

        assert_eq!(all_listener.get_count(), 1);
        assert_eq!(msg_listener.get_count(), 0);
        assert_eq!(field_listener.get_count(), 0);
        assert_eq!(field_value_listener.get_count(), 0);
        assert_eq!(msg_field_value_listener.get_count(), 0);

        info!("filter message");

        assert_eq!(all_listener.get_count(), 2);
        assert_eq!(msg_listener.get_count(), 1);
        assert_eq!(field_listener.get_count(), 0);
        assert_eq!(field_value_listener.get_count(), 0);
        assert_eq!(msg_field_value_listener.get_count(), 0);

        info!(field = 1, "not filter message, field not value");

        assert_eq!(all_listener.get_count(), 3);
        assert_eq!(msg_listener.get_count(), 1);
        assert_eq!(field_listener.get_count(), 1);
        assert_eq!(field_value_listener.get_count(), 0);
        assert_eq!(msg_field_value_listener.get_count(), 0);

        info!(field = 1234, "not filter message, field and value");

        assert_eq!(all_listener.get_count(), 4);
        assert_eq!(msg_listener.get_count(), 1);
        assert_eq!(field_listener.get_count(), 2);
        assert_eq!(field_value_listener.get_count(), 1);
        assert_eq!(msg_field_value_listener.get_count(), 0);

        info!(field = 1234, "filter message");

        assert_eq!(none_listener.get_count(), 0);
        assert_eq!(all_listener.get_count(), 5);
        assert_eq!(msg_listener.get_count(), 2);
        assert_eq!(field_listener.get_count(), 3);
        assert_eq!(field_value_listener.get_count(), 2);
        assert_eq!(msg_field_value_listener.get_count(), 1);
    }

    #[test]
    fn test_filter_fields() {
        let no_fields = EventListener::new(Filters::new());
        let one_field = EventListener::new(Filters::new().field("field_one"));
        let two_fields = EventListener::new(
            Filters::new()
                .field("field_one")
                .field_value("field_two", "123"),
        );
        let subscriber = tracing_subscriber::registry()
            .with(no_fields.clone())
            .with(one_field.clone())
            .with(two_fields.clone());
        let _sub = tracing::subscriber::set_default(subscriber);

        assert_eq!(no_fields.get_count(), 0);
        assert_eq!(one_field.get_count(), 0);
        assert_eq!(two_fields.get_count(), 0);

        info!("no fields message");

        assert_eq!(no_fields.get_count(), 1);
        assert_eq!(one_field.get_count(), 0);
        assert_eq!(two_fields.get_count(), 0);

        info!(field_one = 123, "one field message");

        assert_eq!(no_fields.get_count(), 2);
        assert_eq!(one_field.get_count(), 1);
        assert_eq!(two_fields.get_count(), 0);

        info!(field_one = 123, field_two = 123, "two field message");

        assert_eq!(no_fields.get_count(), 3);
        assert_eq!(one_field.get_count(), 2);
        assert_eq!(two_fields.get_count(), 1);
    }
}
